#include <stdlib.h>
#include <string.h>
#include "expr.h"

enum optype 
{
  opt_bin = (1<<0), opt_pfx = (1<<1), opt_sfx = (1<<2),
  opt_lef = (1<<3), opt_rig = (1<<4),
  opt_par = (1<<5), opt_quo = (1<<6),
};

typedef struct { int op, pos; } oper_t;

typedef struct
{
  char *str;
  int type;    // combination of the opt_* stuff.
               // opt_bin | op_lef  is a left-associative binary operator
               // opt_par | op_rig  is a right parenthesis etc.
  int prec;    // precedence
  int match;   // matching token which closes this parenthesis/quote.
  int display; // number to be set in the returned tree
  unsigned int escape; 
               // I could re-use the 'prec' field, but it's ok.
               // this is the escape character for the corresponding
               // quote mark
} opdef_t;

struct xpp
{
  int cclass[256];
  int s_opd, n_opd; opdef_t *opd;
  int max_oper_len;

  unsigned char *inp, *orig;
  int token;
  char *toktext;
  int tokoper;
  int tokpos;
  int mode; // 'T' or 'O'. expecting a term or an operator the
            // lexer uses this to decide which token to return.

  int n_op,s_op; oper_t *op;
  int n_term, s_term; expr_t **term;
};


enum { TOK_INT, TOK_FLO, TOK_IDENT, TOK_QUOT, 
       TOK_EOF, TOK_ERROR, TOK_OPER };



static int lex(xpp_t* p);
static int lex_ident(xpp_t *p);
static int mk_token(xpp_t *p,int token, unsigned char *end);
static int lex_number(xpp_t *p);
static int mk_error_wpos(xpp_t *p,char *str, int pos);
static int mk_error(xpp_t *p,char *str);
static int new_opdef(xpp_t *p,char* str,int type,int disp);
static int find_opdef(xpp_t *p, char *str,int type, int disp);
static int lex_oper(xpp_t *p);
static int match_text(char *op, char *inp, int len);
static int max_oper_len(xpp_t *p);
static int str_is_prefix(char *str, char *inp);
static int lex_quote(xpp_t *p,int quotop);
static int hexva(unsigned int K);
static void push_op(xpp_t *p, int op);
static void pop_op(xpp_t *p);
static void push_term(xpp_t *p, expr_t *e);
static void pop_term(xpp_t *p);
static expr_t *mkterm(xpp_t *p,int disp);
static expr_t *expr_new(int disp);
static int reduce_op(xpp_t *p, int op,int pos);
static void free_terms(xpp_t *p, int s, int e);
static expr_t *mk_error_expr(xpp_t *p);

static int lex(xpp_t* p)
{
  if (p->toktext) { free(p->toktext); p->toktext= NULL; }
  p->tokoper= 0;
  while(p->cclass[p->inp[0]]==' ') p->inp++;

  switch(p->cclass[p->inp[0]])
  {
  case '!': p->token= TOK_EOF; return 0;
  case 'A': return lex_ident(p);
  case '1': return lex_number(p);
  case '+': return lex_oper(p);
  }
  return mk_error(p, "unknown character");
}

static int lex_ident(xpp_t *p)
{
  unsigned char *end;
  end= p->inp;
  while(p->cclass[end[0]]=='A' || p->cclass[end[0]]=='1') end++;
  return mk_token(p,TOK_IDENT,end);
}

static int mk_token(xpp_t *p,int token, unsigned char *end)
{
  int L;
  if (p->toktext) free(p->toktext);
  L= end - p->inp;
  p->toktext= malloc(L+1);
  p->tokpos= p->inp - p->orig;
  memcpy(p->toktext,p->inp,L);
  p->toktext[L]= 0;
  p->inp= end;
  p->token= token;
  return 0;
}

static int lex_number(xpp_t *p)
{
  unsigned char *end;
  end= p->inp;
  while(p->cclass[end[0]]=='1') end++;
  if (end[0]!='.') return mk_token(p,TOK_INT,end);
  end++;
  while(p->cclass[end[0]]=='1') end++;
  if (end[0]!='E' && end[0]!='e') return mk_token(p,TOK_FLO,end);
  end++;
  if (end[0]=='+' || end[0]=='-') end++;
  if (p->cclass[end[0]]!='1') return mk_error(p, "bad number token");
  while(p->cclass[end[0]]=='1') end++;
  return mk_token(p,TOK_FLO,end);
}

static int mk_error_wpos(xpp_t *p,char *str, int pos)
{
  p->token= TOK_ERROR;
  p->tokpos= pos;
  if (p->toktext) free(p->toktext);
  p->toktext= strdup(str);
  p->tokoper= 0;
  return 1;
}

static int mk_error(xpp_t *p,char *str)
{
  return mk_error_wpos(p,str, p->inp-p->orig);
}


static int new_opdef(xpp_t *p,char* str,int type,int disp)
{
  int R;
   // only the first character is classified this way, that's all
   // we need.
  p->cclass[*(unsigned char*)str]='+';
  if (p->n_opd==p->s_opd)
  {
    p->s_opd+= p->s_opd/2+10;
    p->opd= realloc(p->opd,sizeof(p->opd[0])*p->s_opd);
  }
  memset(p->opd+p->n_opd,0, sizeof(p->opd[0]));
  R= p->n_opd;
  p->n_opd++;
  p->opd[R].str= strdup(str);
  p->opd[R].type= type;
  p->opd[R].display= disp;
  return R; 
}

static int find_opdef(xpp_t *p, char *str,int type, int disp)
{
  int i;
  for(i=0;i<p->n_opd;i++)
    if (!strcmp(p->opd[i].str,str)) return i;
  return new_opdef(p,str,type,disp);
}

static int lex_oper(xpp_t *p)
{
  int max_oper,input_len,match_len,best_match_len;
  int matches[10];
  int nmatches; 
  int i,R,j;

  max_oper= max_oper_len(p);
  for(input_len=0;p->inp[input_len] && input_len<max_oper;input_len++) ;

  nmatches= 0;
  best_match_len= 0;

  for(i=0;i<p->n_opd;i++)
  {
    match_len= match_text(p->opd[i].str, (char*) p->inp, input_len);
    if (match_len==0) continue;
    if (match_len>best_match_len)
    {
      matches[0]= i;
      nmatches= 1;
      best_match_len= match_len;
    }
    else if (match_len==best_match_len)
    {
      matches[nmatches++]= i;
    }
  }

  if (best_match_len==0) goto error;

  if (p->mode=='T') // was expecting a term..
  {
    for(j=0;j<nmatches;j++)
    {
      int T;
      i= matches[j];
      T= p->opd[i].type;
                    // we can return a prefix operator 
                    // or a normal parenthesis
      if ((T&opt_pfx) || ((T&opt_par) && !(T&opt_sfx)))
      {
        R= i; goto done;
      }
      if ((T&opt_quo) && (T&opt_lef))
      {
        return lex_quote(p, i);
      }
    }
  }
  else
  for(j=0;j<nmatches;j++)
  {
     int T;
     i= matches[j];
     T= p->opd[i].type;
     if (T&opt_sfx || T&opt_bin || ((T&opt_par) && (T&opt_rig)))
     {
       R= i; goto done;
     }
  }
error:
  return mk_error(p, "unknown operator");
done:
  p->tokoper= R;
  return mk_token(p, TOK_OPER, p->inp+best_match_len);
}

static int match_text(char *op, char *inp, int len)
{
  int ol,i;
  ol= strlen(op);
  if (ol>len) return 0;
  for(i=0;i<ol;i++) if (op[i]!=inp[i]) return 0;
  return ol;
}

static int max_oper_len(xpp_t *p)
{
  int i,L,m;
  if (p->max_oper_len) return p->max_oper_len;
  m= 0;
  for(i=0;i<p->n_opd;i++)
  {
    L= strlen(p->opd[i].str);
    if (L>m) m= L;
  }
  p->max_oper_len= m;
  return m;
}

static int str_is_prefix(char *str, char *inp)
{
  int i;
  for(i=0;str[i];i++) if (str[i]!=inp[i]) return 0;
  return 1;
}

static int lex_quote(xpp_t *p,int quotop)
{
  unsigned char *s,*e;
  opdef_t *O,*C;
  int i,j, L, hv1, hv2;
  char *R;

  R= NULL;
  O= p->opd + quotop;
  C= p->opd + O->match;
  s=e= p->inp+strlen(O->str);

  p->tokpos= p->inp - p->orig;

  while(1)
  {
    if (!e[0]) goto E_eof;
    if (e[0]==O->escape)
    {
      e++;
      if (!e[0]) goto E_eof;
      e++;
      continue;
    }
    if (str_is_prefix(C->str,(char*)e)) break;
    e++;
  }
  L= e-s;
  R= malloc(L+1);
  j= 0;
  for(i=0;i<L;i++)
  {
    if (s[i]!=O->escape) { R[j++]= s[i]; continue; }
    i++; 
    if (i==L) goto E_bad_escape;
    switch(s[i])
    {
    case 'n': R[j++]= '\n'; break;
    case 't': R[j++]= '\t'; break;
    case 'x': 
      if (L-i<3) goto E_bad_escape;
      hv1= hexva(s[i+1]);
      hv2= hexva(s[i+2]);
      if (hv1==-1 || hv2==-1) goto E_bad_escape;
      R[j++]= (hv1<<4) | hv2;
      i+= 2;
      break;
    default:
      R[j++]= s[i];
      break;
    }
  }
  R[j]= 0;
  p->token= TOK_QUOT;
  if (p->toktext) free(p->toktext);
  p->toktext= R;
  p->tokoper= quotop;

  p->inp= e + strlen(C->str);
  return 0;

E_eof:
  return mk_error(p, "unmatched quote");
E_bad_escape:
  if (R) free(R);
  p->inp= s+i;
  return mk_error(p,"bad escape sequence");
}

static int hexva(unsigned int K)
{
  if (K>='a' && K<='f') return K-'a'+10;
  if (K>='A' && K<='F') return K-'A'+10;
  if (K>='0' && K<='9') return K-'0';
  return -1;
}



static void push_op(xpp_t *p, int op)
{
  if (p->n_op==p->s_op)
  {
    p->s_op+= p->s_op/2 + 10;
    p->op= realloc(p->op, sizeof(p->op[0])*p->s_op);
  }
  p->op[p->n_op].op= op;
  p->op[p->n_op].pos= p->tokpos;
  p->n_op++;
}

static void pop_op(xpp_t *p)
{
  p->n_op--;
}

static void push_term(xpp_t *p, expr_t *e)
{
  if (p->n_term==p->s_term)
  {
    p->s_term+= p->s_term/2+10;
    p->term= realloc(p->term, sizeof(p->term[0])*p->s_term);
  }
  p->term[p->n_term++]= e;
}

static void pop_term(xpp_t *p)
{
  p->n_term--;
}


static expr_t *mkterm(xpp_t *p,int disp)
{
  expr_t *E;
  E= expr_new(disp);
  E->data= p->toktext;
  p->toktext= NULL;
  E->pos= p->tokpos;
  return E;
}

static expr_t *expr_new(int disp)
{
  expr_t *E;
  E= calloc(1,sizeof(*E));
  E->op= disp;
  return E;
}

static int reduce_op(xpp_t *p, int op,int pos)
{
  opdef_t *Y;
  expr_t *A[2],*E;
  int nA;
  int tY;
  int i;
  Y= p->opd + op;
  tY= Y->type;

  A[1]= NULL;
  if ((tY&opt_bin) || ((tY&opt_par)&&(tY&opt_sfx))) nA= 2;
                                           else nA= 1;

  if (p->n_term<nA) goto underflow;
  for(i=nA-1;i>=0;i--) { A[i]= p->term[p->n_term-1]; pop_term(p); }

  E= expr_new(Y->display);
  E->left= A[0];
  E->right= A[1];
  E->pos= pos;
  
  push_term(p,E);
  return 0;
underflow:
  mk_error_wpos(p,"internal error: stack underflow",pos); 
  if (!p->term) { p->term= malloc(sizeof(p->term[0])); p->s_term= 1; }
  p->term[0]= mk_error_expr(p);
  return 1;
}



static void free_terms(xpp_t *p, int s, int e)
{
  int i;
  for(i=s;i<=e;i++) xpp_free_expr(p->term[i]);
}

static expr_t *mk_error_expr(xpp_t *p)
{
  expr_t *E;
  E= expr_new(XPP_ERROR);
  E->pos= p->tokpos;
  E->data= p->toktext;
  p->toktext= NULL;

  free_terms(p, 0, p->n_term-1);
  p->n_term= 0;
  p->n_op= 0;
  return E;
}


_XPP_PROTO xpp_t *xpp_new()
{
  int i;
  xpp_t *p;
  p= calloc(1,sizeof(*p));
  for(i=1;i<=32;i++) p->cclass[i]= ' ';
  p->cclass[127]= ' ';
  p->cclass[0]  = '!';
  for(i='a';i<='z';i++) p->cclass[i]= 'A';
  for(i='A';i<='Z';i++) p->cclass[i]= 'A';
  p->cclass['_']= 'A'; // until overwritten
  for(i='0';i<='9';i++) p->cclass[i]= '1';

  return p;
}



_XPP_PROTO void xpp_binary_operator
      (xpp_t* p,char *str,int disp,int prec, int assoc)
{
  int Y;
  Y= new_opdef(p,str, opt_bin | (assoc=='L' ? opt_lef : opt_rig), disp);
  p->opd[Y].prec= prec;
}

_XPP_PROTO void xpp_prefix_operator
      (xpp_t* p,char *str,int disp)
{
  new_opdef(p,str, opt_pfx, disp);
}

_XPP_PROTO void xpp_postfix_operator
      (xpp_t *p,char *str,int disp)
{
  new_opdef(p,str, opt_sfx, disp);
}

_XPP_PROTO void xpp_parenthesis
      (xpp_t *p,char *left,char *right,int disp)
{
  int od1,od2;
  od1= new_opdef(p,left, opt_lef | opt_par, disp);
  od2= find_opdef(p,right,opt_rig | opt_par, disp);
  p->opd[od1].match= od2;
}

_XPP_PROTO void xpp_postfix_parenthesis
      (xpp_t *p,char *left,char *right,int disp)
{
  int od1,od2;
  od1= new_opdef(p,left, opt_lef | opt_par | opt_sfx, disp);
  od2= find_opdef(p,right,opt_rig | opt_par, disp);
  p->opd[od1].match= od2;
}


_XPP_PROTO void xpp_quote
      (xpp_t *p, char *left, char *right, int escape, int disp)
{
  int od1,od2;
  od1= new_opdef(p, left, opt_quo | opt_lef, disp);
  od2= find_opdef(p,right,opt_quo | opt_rig, disp);
  p->opd[od1].match= od2;
  p->opd[od1].escape= escape;
}


_XPP_PROTO expr_t* xpp_run(xpp_t *p,char *input)
{
   opdef_t *T;
   int oT, K, Kpos;
   p->orig= p->inp= (unsigned char *)input;

P_term:
   p->mode= 'T';
   lex(p);
   switch(p->token)
   {
   case TOK_EOF: goto E_expect_term;
   case TOK_ERROR: goto E_lexer;
   case TOK_INT: push_term(p, mkterm(p,XPP_INT)); goto P_op;
   case TOK_FLO: push_term(p, mkterm(p,XPP_FLO)); goto P_op;
   case TOK_IDENT: push_term(p,mkterm(p,XPP_IDENT)); goto P_op;
   case TOK_QUOT: push_term(p, mkterm(p,p->opd[p->tokoper].display));
                  goto P_op;
   case TOK_OPER: break;
   default: goto E_lexer;
   }

   oT= p->opd[p->tokoper].type;

   if (oT&opt_pfx)
   {
     push_op(p,p->tokoper);
     goto P_term;
   }

   if ((oT&opt_par) && (oT&opt_lef) && !(oT&opt_sfx))
   {
     push_op(p,p->tokoper);
     goto P_term;
   }

   goto E_expect_term;

P_op:
   p->mode= 'O';
   lex(p);
   switch(p->token)
   {
   case TOK_ERROR: goto E_lexer;
   case TOK_EOF: goto finish_off_eof;
   case TOK_QUOT:
   case TOK_IDENT:
   case TOK_FLO:
   case TOK_INT: goto E_expect_op;
   case TOK_OPER: break;
   default: goto E_lexer;
   }
   oT= p->opd[p->tokoper].type;
   if ((oT&opt_par) && (oT&opt_rig)) goto finish_off;

   if ((oT&opt_par) && (oT&opt_lef) && (oT&opt_sfx))
   {
     push_op(p, p->tokoper);
     goto P_term;
   }
   if (oT&opt_sfx)
   {
     if (reduce_op(p, p->tokoper, p->tokpos)) return p->term[0];
     goto P_op;
   }
   if (oT&opt_bin)
   {
     opdef_t *C;
     C= p->opd + p->tokoper;
     while(p->n_op)
     {
       K= p->op[p->n_op-1].op;
       Kpos= p->op[p->n_op-1].pos;
       T= p->opd + K;
       if ((T->type&opt_pfx)) 
       { if (reduce_op(p, K, Kpos)) return p->term[0]; 
         pop_op(p); continue; }
       if ((T->type&opt_bin)==0) break; // this can happen if there was
                                        // a parenthesis on the stack
       if (T->prec<C->prec) break;
       if (T->prec>C->prec || T->type&opt_lef) 
            { if (reduce_op(p,K,Kpos)) return p->term[0];
              pop_op(p); continue; }
       break;
     }
     push_op(p,p->tokoper);
     goto P_term;
   }
   goto E_expect_op;

E_expect_term:
   mk_error_wpos(p, "a term was expected", p->tokpos);
   return mk_error_expr(p);

E_lexer:
   return mk_error_expr(p);

E_expect_op:
   mk_error_wpos(p, "an operator was expected", p->tokpos);
   return mk_error_expr(p);

E_unmatched_open:
   mk_error_wpos(p,"unmatched open parenthesis", Kpos);
   return mk_error_expr(p);

finish_off_eof:
   while(p->n_op)
   {
     K= p->op[p->n_op-1].op;
     Kpos= p->op[p->n_op-1].pos;
     T= p->opd + K;
     if ((T->type&opt_par)) goto E_unmatched_open;
     if (reduce_op(p, K, Kpos)) return p->term[0];
     pop_op(p); 
   }
   if (p->n_term>1)
   { 
     // should never happen!! maybe return an error?? 
     free_terms(p,1,p->n_term-1);
   }
   p->n_term= 0;
   return p->term[0];
finish_off:
   while(p->n_op)
   {
     K= p->op[p->n_op-1].op;
     Kpos= p->op[p->n_op-1].pos;
     T= p->opd + K;
     if ((T->type&opt_par)) 
     {
       if (T->match!=p->tokoper) goto E_unmatched_open;
       if (reduce_op(p, K,Kpos)) return p->term[0];
       pop_op(p);
       break;
     }
     if (reduce_op(p, K,Kpos)) return p->term[0];
     pop_op(p); 
   }
   goto P_op;
}


_XPP_PROTO void xpp_free_expr(expr_t *expr)
{
  if (expr->data) free(expr->data);
  if (expr->left) xpp_free_expr(expr->left);
  if (expr->right) xpp_free_expr(expr->right);
  free(expr);
}

_XPP_PROTO void xpp_free(xpp_t *p)
{
  int i;
  if (p->toktext) free(p->toktext);
  for(i=0;i<p->n_opd;i++) free(p->opd[i].str);
  if (p->opd) free(p->opd);
  if (p->op) free(p->op);
  if (p->term) free(p->term);
  free(p);
}

