#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <stdio.h>
#include "strbuf.h"
#include "expr.h"
#include "widget.h"
#include "bytecode.h"
#include "n-eval.h"
#include "critbit.h"

typedef struct { char *str; int iv; } strintpair_t;

typedef struct
{
  int type;    // 'W'= widget coord, 'V'= local var, 'G'= global var
  int visit;           // flag to exclude infinite recursion

  expr_t *expr;
  bc_buffer_t *access; // code to access this thing. this is non-null
                       // if the computation has been done successfully.
  Uwidget_t *widget;
  int wnum;
  int coord_idx; // 0..5 for widget coord

  Uvar_t *var;
  int localnum;
  int globalnum;  // these two could be combined, since the type field
                  // already differentiates between locals and globals.
} value_t;

typedef struct
{
  Uwidget_t *widget;
  char *ident;
  int wnum; // within the screen
  value_t *x,*y,*w,*h,*bot,*rig;
} widget_eval_t;

typedef struct
{
  int type;  // 'W'idget 'L'ocal 'G'lobal
             // if type is 'L' or 'G', ptr points to 
            //   a value_t, otherwise, it points to a
             // widget_eval_t
  void *ptr;
} symbol_t;

static strmap_t *symtab;

static int n_locals;
static int n_widgets;
static widget_eval_t *wevals;

void evaldis_coords(Uscreen_t *S,int sw,int sh);
char *eval_screen_dim(Uscreen_t *S);


static int coord_accessor(int csel)
{
  switch(csel)
  {
  case wTOP: return GUII_WLAY_Y;
  case wLEF: return GUII_WLAY_X;
  case wWID: return GUII_WLAY_W;
  case wHEI: return GUII_WLAY_H;
  }
  return GUII_WLAY_Y;
}

bc_buffer_t* bc_buffer_new(int size)
{
  bc_buffer_t *B;
  B= malloc(sizeof(*B));
  B->size= size;
  B->len= 0;
  B->data= malloc(size);
  return B;
}


void bc_buffer_put(bc_buffer_t *buf, unsigned char V)
{
  if (buf->len==buf->size)
  {
    buf->size += buf->size/2 + 20;
    buf->data= realloc(buf->data, buf->size);
  }
  buf->data[buf->len++]= V;
}

#define put_byte   bc_buffer_put
#define put_code   put_byte

static int bc_buffer_tell(bc_buffer_t *buf)
{
  return buf->len;
}

static void bc_buffer_rewind(bc_buffer_t *buf, int pos)
{
  buf->len= pos;
}

void bc_buffer_append(bc_buffer_t *dst, bc_buffer_t *src)
{
  if (dst->len+src->len>dst->size)
  {
     dst->size= dst->len+src->len+dst->size/4;
     dst->data= realloc(dst->data, dst->size);
  }
  memcpy(dst->data+dst->len, src->data, src->len);
  dst->len+= src->len;
}

static int alloc_local_var()
{
  return n_locals++;
}

static expr_t *make_widget_expression(Uwidget_t *W, char *f1, int op, char *f2)
{
  char buf[200];
  expr_t *e;
  sprintf(buf,"%s.%s %c %s.%s",W->ident,f1, op, W->ident, f2);
  e= xpp_run(xpp, buf);
  if (!e || e->op==XPP_ERROR) e= xpp_run(xpp, "0");
  return e;
}

static value_t *value_new(int type)
{
  value_t *V;
  V= calloc(1,sizeof(*V));
  V->type= type;
  return V;
}


static value_t* coord_value
  (Uwidget_t *W, int wnum, int coord_idx, 
   expr_t *e, char *altF1, int op, char *altF2)
{
  value_t *V;
  if (!e) e= make_widget_expression(W, altF1, op, altF2);

  V= value_new('W');
  V->expr= e;

  V->widget= W;
  V->wnum= wnum;
  V->coord_idx= coord_idx;

  return V;
}

static value_t* coord_value_as_local
  (Uwidget_t *W, int wnum, int coord_idx,
   expr_t *e, char *altF1, int op, char *altF2)
{
  value_t *V;
  if (!e) e= make_widget_expression(W, altF1, op, altF2);

  V= value_new('V');
  V->expr= e;
  V->localnum= alloc_local_var();

  V->wnum= wnum;
  V->widget= W;
  V->coord_idx= coord_idx;

  return  V;
}
   

static void widget_to_weval(Uwidget_t *W,int wnum,widget_eval_t *L)
{
  L->x= coord_value(W, wnum, wLEF, W->pos.E[wLEF],  "right", '-', "width" );
  L->y= coord_value(W, wnum, wTOP, W->pos.E[wTOP],  "bottom", '-', "height" );
  L->w= coord_value(W, wnum, wWID, W->pos.E[wWID],  "right", '-', "left" );
  L->h= coord_value(W, wnum, wHEI, W->pos.E[wHEI],  "bottom", '-',  "top" );

  L->bot= coord_value_as_local(W, wnum, wBOT, W->pos.E[wBOT],
                                                    "top", '+', "height" );
  L->rig= coord_value_as_local(W, wnum, wRIG, W->pos.E[wRIG],
                                                    "left", '+', "width" );
  L->wnum= wnum;
  L->widget= W;
  L->ident= W->ident;
}

static void symtab_insert(char *name, int type, void *ptr)
{
  symbol_t *S;
  S= malloc(sizeof(*S));
  S->type= type;
  S->ptr= ptr;
  strmap_insert(symtab, name, S);
}

static void make_wevals(Uscreen_t *S)
{
  Uwidget_t *W;
  int N;
  for(W=S->widgets,N=0;W;W=W->next,N++) ;
  n_widgets= N;
  wevals= malloc(sizeof(wevals[0])*N);
  for(W=S->widgets,N=0;W;W=W->next,N++)
  {
     widget_to_weval(W, N, wevals+N);
     symtab_insert(W->ident, 'W', wevals+N);
  }
}



static void symtab_insert_locals(Uscreen_t *S)
{
  Uvar_t *var;
  value_t *V;
  for(var=S->vars;var;var=var->next)
  {
    V= value_new('V');
    V->var= var;
    V->expr= var->expr;
    V->localnum= alloc_local_var();
    symtab_insert(var->ident, 'V', V);
  }
}

static bc_buffer_t *cbuf;

static value_t *error_value;
static char *error_message;

static int err1arg(char *msg, char *arg)
{
  char *R;
  R= malloc(strlen(msg)+strlen(arg)+1);
  sprintf(R, "%s%s",msg,arg);
  error_message= R;
  return 1;
}


static int compute_value(value_t *V);

static int compute_dot_expr(expr_t *L, expr_t *R)
{
  widget_eval_t *WE;
  value_t *V;
  symbol_t *SS;

  if (L->op!=XPP_IDENT || R->op!=XPP_IDENT) return 1;
  SS= strmap_find(symtab, L->data);
  if (!SS)  return err1arg("unknown widget ", L->data);
  if (SS->type!='W') return err1arg(L->data, "is not a widget");
  WE= SS->ptr;
  
  if (!strcmp(R->data,"left")) V= WE->x;
  else if (!strcmp(R->data,"right")) V= WE->rig;
  else if (!strcmp(R->data,"top")) V= WE->y;
  else if (!strcmp(R->data,"bottom")) V= WE->bot;
  else if (!strcmp(R->data, "width")) V= WE->w;
  else if (!strcmp(R->data, "height")) V= WE->h;
  else return err1arg("unknown field ",R->data);

  if (compute_value(V)) return 1;
  bc_buffer_append(cbuf, V->access);
  return 0;
}

static int compute_ident(char *name)
{
  value_t *V;
  symbol_t *SS;
  SS= strmap_find(symtab, name);
  if (!SS) return err1arg("unknown variable ",name);
  if (SS->type!='V' && SS->type!='G') 
    return err1arg(name, " is not a variable");
  V= SS->ptr;
  if (compute_value(V)) return 1; 
  bc_buffer_append(cbuf, V->access);
  return 0;
}

static void put_int(bc_buffer_t *B, uint32_t V)
{
  if (V&(0xff<<24))
  {
    put_byte(B, GUII_C4);
    put_byte(B, V>>24);
    put_byte(B, V>>16);
    put_byte(B, V>>8);
    put_byte(B, V);
  }
  else if (V&(0xff<<16))
  {
    put_byte(B, GUII_C3);
    put_byte(B, V>>16);
    put_byte(B, V>>8);
    put_byte(B, V);
  }
  else if (V&(0xff<<8))
  {
    put_byte(B, GUII_C2);
    put_byte(B, V>>8);
    put_byte(B, V);
  }
  else if ( V & 0x80 )
  {
    put_byte(B, GUII_C1);
    put_byte(B, V);
  }
  else put_byte(B, 0x80 | V );
}

void bc_buffer_put_int(bc_buffer_t *B, uint32_t V) { put_int(B,V); }

static void put_flo(bc_buffer_t *B, float F)
{
  union { float F; uint32_t I; } D;
  uint32_t V;
  D.F= F;
  V= D.I;

  put_byte(B, GUII_CF);
  put_byte(B, V>>24);
  put_byte(B, V>>16);
  put_byte(B, V>>8);
  put_byte(B, V);
}

void bc_buffer_put_flo(bc_buffer_t *B, float F) { put_flo(B,F); }

static int compute_expr(expr_t *E);

static int compute_binop(expr_t *E, unsigned char op)
{
  if (compute_expr(E->left)) return 1;
  if (compute_expr(E->right)) return 1;
  put_code(cbuf,op);
  return 0;
}

static int compute_expr(expr_t *E)
{
  int i;
  switch(E->op)
  {
   case EXP_ADD: return compute_binop(E, GUII_ADD);
   case EXP_SUB: return compute_binop(E, GUII_SUB);
   case EXP_MUL: return compute_binop(E, GUII_MUL);
   case EXP_DIV: return compute_binop(E, GUII_DIV);
   case EXP_MOD: return compute_binop(E, GUII_MOD);
   case EXP_NEG:
     put_int(cbuf, 0); if (compute_expr(E->right)) return 1;
     put_code(cbuf, GUII_SUB);
     return 0;
   case EXP_PLUS: return compute_expr(E->right);
   case EXP_PAREN_R: return compute_expr(E->left);
   case XPP_FLO: put_flo(cbuf, strtof(E->data, NULL)); return 0;
   case XPP_INT: put_int(cbuf, strtol(E->data,NULL,0)); return 0;
   case XPP_ERROR: return 1;
   case XPP_IDENT:
     return compute_ident(E->data);
   case EXP_DOT:
     return compute_dot_expr(E->left, E->right);
  }
  return 1;
}


static int compute_value(value_t *V)
{
  int R;
  if (V->access) return 0; // already known
  if (V->visit) 
  {
    error_value= V;
    error_message= "A circular computation is found";
    return 1;
  }
  V->visit= 1;
  R= compute_expr(V->expr);
  V->visit= 0;
  if (R) 
  {
    if (!error_value)  error_value= V; 
    if (!error_message)  error_message= "Can not compute"; 
    return 1;
  }
  V->access= bc_buffer_new(50);
  // After compute_expr finishes, the code buffer contains code
  // which would put the result of V->expr at the top of the stack.
  // Now, we generate code which would store that value at the proper
  // place in the Wcoor array.
  if (V->type=='W')
  {
     put_int(cbuf, V->wnum);
     put_code(cbuf, coord_accessor(V->coord_idx));
     put_code(cbuf, GUII_STOREINT);

     put_int(V->access, V->wnum);
     put_code(V->access, coord_accessor(V->coord_idx));
     put_code(V->access, GUII_LOADINT);
  }
  else
  {
     put_int(cbuf, V->localnum);
     put_code(cbuf, GUII_LOCAL);
     put_code(cbuf, GUII_STOREINT);

     put_int(V->access, V->localnum);
     put_code(V->access, GUII_LOCAL);
     put_code(V->access, GUII_LOADINT);
  }
  // FIXME: handle globals
  return 0;
}

static int compute_widget_coords(widget_eval_t *L)
{
  if (compute_value(L->x) 
     || compute_value(L->y) 
     || compute_value(L->w)
     || compute_value(L->h))
   return 1;
  return 0;  
}


void bc_buffer_free(bc_buffer_t *buf)
{
  free(buf->data);
  free(buf);
}

static bc_buffer_t *access_var(int type, int num)
{
  bc_buffer_t *B;
  B= bc_buffer_new(10);
  put_int(B, num);
  put_code(B, type=='V' ? GUII_LOCAL : GUII_GLOBAL );
  put_code(B, GUII_LOADINT);
  return B;
}


static void symtab_insert_predefs()
{
  value_t *V;
  int i;
  //
  // The variables inserted here have no corresponding Uvar_t
  // since they are not true variables. These are in fact
  // constants which the user can refer to in his expressions
  // So, no expr is given. Just an access buffer.
  //

  static const strintpair_t plocals[]=
  {
    { "WIDTH", GUILV_WIDTH },
    { "HEIGHT", GUILV_HEIGHT },
    { NULL, 0 }
  };
  static const strintpair_t pglobals[]=
  {
    { "font1height", GUIGV_FONT1HEIGHT },
    { "font2height", GUIGV_FONT2HEIGHT },
    { "font3height", GUIGV_FONT3HEIGHT },
    { "font4height", GUIGV_FONT4HEIGHT },
    { "font5height", GUIGV_FONT5HEIGHT },
    { "gui_scrollbar_width", GUIGV_GUI_SCROLLBAR_WIDTH },
    { "gui_scrollbar_height", GUIGV_GUI_SCROLLBAR_HEIGHT },
    { "gui_sysbar_height", GUIGV_GUI_SYSBAR_HEIGHT },
    { NULL, 0 }
  };

  for(i=0;plocals[i].str;i++)
  {
    V= value_new('V');
    V->localnum= plocals[i].iv;
    V->access= access_var('V', plocals[i].iv);
    symtab_insert(plocals[i].str, 'V', V);
  }

  for(i=0;pglobals[i].str;i++)
  {
    V= value_new('G');
    V->globalnum= pglobals[i].iv;
    V->access= access_var('G', pglobals[i].iv);
    symtab_insert(pglobals[i].str, 'G', V);
  }
}

typedef struct
{
  strbuf_t *B;
  int linelen;
} Gstrbuf_t;

static Gstrbuf_t *Gstrbuf_new()
{
  Gstrbuf_t *G;
  G= malloc(sizeof(*G));
  G->B= strbuf_init();
  G->linelen= 0;
  return G;
}

static void GputS(Gstrbuf_t *G, char *K)
{
  int L;
  L= strlen(K);
  if (G->linelen+L>72)
  {
    strbuf_puts_l(G->B, "\n", 1);
    G->linelen= 0;
  }
  strbuf_puts_l(G->B, K, L);
  G->linelen+= L;
}

static void GputB(Gstrbuf_t *G,unsigned char X)
{
  static char buf[20];
  sprintf(buf,"%d,",X);
  GputS(G,buf);
}

static void GputI(Gstrbuf_t *G,int X)
{
  static char buf[20];
  sprintf(buf,"%d,",X);
  GputS(G,buf);
}

static char *Gfree(Gstrbuf_t *G)
{
  char *R;
  R= strbuf_collect(G->B);
  free(G);  
  return R;
}


char *bc_buffer_to_string(bc_buffer_t *cb, int n_loc,int only_code)
{
  int i;
  Gstrbuf_t *G;
  G= Gstrbuf_new();
  if (!only_code)
  {
    GputS(G, " .n_locals= ");
    GputI(G, n_loc);
    GputS(G, " .length= ");
    GputI(G, cb->len);
    GputS(G, " .data= (unsigned char[]) {");
  }
  for(i=0;i<cb->len;i++) GputB(G, cb->data[i]);
  if (!only_code) GputS(G, "}");
  return Gfree(G);
}

bc_buffer_t *compile_widget_layout(Uscreen_t *S,int *n_loc)
{
  int nw;
  int i;

  symtab= strmap_new();
  symtab_insert_predefs();
  n_locals= n_predef_locals;
  symtab_insert_locals(S);

  make_wevals(S);
  cbuf= bc_buffer_new(256);
 
  for(i=0;i<n_widgets;i++)
    if (compute_widget_coords(wevals+i)) goto fail;

back_from_fail:
  //
  // there is some memory leak due to expressions that were created
  // by us (ie. u.right= u.left+u.width etc). I'm going to ignore them.
  //
  free(wevals);
  *n_loc= n_locals;

  return cbuf;

fail:
  bc_buffer_free(cbuf);
  cbuf= NULL;
  goto back_from_fail;
}

void print_bytecode(char *fname, bc_buffer_t *B)
{
  FILE *f;
  int i;
  char *K;
  f= fopen(fname, "w");
  if (!f) return ;
  K= bc_buffer_to_string(B, n_locals,0);
  fputs(K,f);
  fclose(f);
  free(K);
}

void bc_buffer_put_str(bc_buffer_t *B, char* str)
{
  int L,i;
  if (!str) str= "";
  L= strlen(str);
  put_int(B,L+1);
  put_code(B, GUII_STR);
  for(i=0;i<=L;i++) put_byte(B, str[i]);
}

static void compile_screen_dimension(expr_t *E, unsigned char inst)
{
  if (!E) return ;
  if (compute_expr(E)) return ;
  bc_buffer_put_int(cbuf, 0);
  bc_buffer_put(cbuf, inst );
  bc_buffer_put(cbuf, GUII_STOREINT);
}

// Compiling a screen layout is much simpler than compiling a widget
// layout. No expression can refer to the computed dimensions (i.e.
// the expression for width can't make use of the height value etc.)
// so we don't need to store these in a table or something. Also,
// the expressions themselves can't refer to widgets. These expressions
// are basic arithmetic expressions or even simple numbers. Not all the
// dimensions need to be given because there are default values for
// each dimension.
bc_buffer_t *compile_screen_layout(Uscreen_t *S,int *Rnlocals)
{
  symtab= strmap_new();
  symtab_insert_predefs();
  n_locals= n_predef_locals;
  symtab_insert_locals(S);

  cbuf= bc_buffer_new(256);
 
  // FIXME: handle errors..
  compile_screen_dimension(S->pos.E[wLEF], GUII_WLAY_X);
  compile_screen_dimension(S->pos.E[wTOP], GUII_WLAY_Y);
  compile_screen_dimension(S->pos.E[wWID], GUII_WLAY_W);
  compile_screen_dimension(S->pos.E[wHEI], GUII_WLAY_H);
    // if none of the above values generate an opcode, I will
    // so that we don't have to deal with an empty buffer.
  bc_buffer_put(cbuf, GUII_NOP);
  *Rnlocals= n_locals;
  return cbuf;
}
