

#define GFX_UPDATE (1<<0)
#define GFX_ALPHA  (1<<1)
#define GFX_BACKGND (1<<2)   // use background color instead of pixel buffer


/***********************************************************************
 *                                                                     *
 *                                TYPES                                *
 *                                                                     *
 ***********************************************************************/
#ifdef GFX_X11
/***********************************************************************
  Images can be shared on non-shared. They are represented by the
  following struct.
 ***********************************************************************/
typedef struct 
{
  uint32_t *data;
  unsigned int width,height, bpl;
  int shared;
  XShmSegmentInfo shminfo;
  XImage *x11image;
} ximage_t;

#endif

#ifdef GFX_ANDROID
typedef struct req request_t;
typedef struct app app_t;

/***********************************************************************
 * #A.APP                                                              *
 *                 Android Application Object                          *
 *                                                                     *
 ***********************************************************************
  The application object is called an 'instance' in the android documents.
  It contains the necessary data to properly handle and serialize events.
  Window handle etc. are also contained in this object.
 ***********************************************************************/

struct app {
      // command pipe. This is used for reporting system events such
      // as Pause, Resume, etc.
  int cmdpipe_r, cmdpipe_w;

      // The worker thread is actually the 'main' thread for our
      // application. It initializes the input system and then calls
      // android_main. When that function is done, it rips up the
      // input/output objects and quits.
  pthread_t worker_thread;

      // These two are used for signalling various conditions. 
  pthread_mutex_t mutex;
  pthread_cond_t cond;

      // This is one of the conditions. When this is set to 1, 
      // the thread which fired off the worker thread can continue
      // in its Java world.
  int running;

  void *init_state; size_t init_state_size;

  ANativeActivity *activity;
  ALooper *looper;
  AConfiguration *config;
  AInputQueue *inputq;
  ANativeWindow *window;

  request_t *quit_request;
  request_t *save_request;
  request_t *stop_request;
  int last_pointer_id;
};

static app_t *my_app;
#endif

typedef struct { int w,h,stride; uint8_t *data; } graypix_t;
typedef struct { int w,h,stride; uint32_t *data; } colorpix_t;

#define PIX(A,x,y)  (A).data[ (x) + (y)*(A).stride]

typedef int (*event_function_t)(void *,gfx_event_t*);


/***********************************************************************
  Each widget is stored in a widget object. Normally, the individual
  widgets store their information in their own structs, but they are
  declared such that they can be safely cast to widget_t.
 ***********************************************************************/
typedef struct widget
{
  int x,y,w,h;
  int klass;
  int number;   // this is for accessing the array from the pointer

  char can_focus, has_focus;
  void (*focus)(void *,int);
  int  (*handle_key)(void *,gfx_event_t *);

  int shown;
  void (*show)(struct widget *);
  void (*hide)(struct widget *);
  void (*position)(struct widget *,int ,int ,int ,int ); // x y w h
  void (*destroy)(struct widget *);
} widget_t;

typedef struct
{
  int ident;
  int n_widgets,s_widgets;
  widget_t **widgets;
} radio_group_t;

typedef struct
{
  int  x,y,w,h;
  int  (*func)(void *,gfx_event_t*);
  void *arg;
  int  *enabled;
  int button;
} click_handler_t;

typedef struct
{
  int ident;
  unsigned int value;
  int (*func)(void *,gfx_event_t *);
  void *arg;
} gui_timer_t;

/***********************************************************************
  A screen contains the widget list as widget numbers. Other than that,
  it contains the usual stuff such as mouse click handlers or grab
  handlers etc.
 ***********************************************************************/
typedef struct screen
{
  int n_widgets;
  int *widgets;

  int n_click_handlers, s_click_handlers;
  click_handler_t *click_handlers;
  click_handler_t release_handler;
  click_handler_t motion_handler;
  click_handler_t grabber;

  widget_t *focus_widget;
  widget_t *focus_clicked;

  struct screen *up, *down;
  uint32_t *backstore;
  uint32_t back_w, back_h;

  int x,y,w,h; 

  int n_rgroups, s_rgroups;
  radio_group_t *rgroups;

  int n_timers,s_timers;
  gui_timer_t *timers;

  int visible_parts_time;
  int n_visible_parts;
  rect_t *visible_parts;
} screen_t;

static screen_t *screen; // top of the stack

/***********************************************************************
  The visible parts list becomes invalid whenever a new screen is entered
  or the current screen is left. I keep track of this by counting the
  number of times enter_screen and leave_screen functions have been
  called. When there is no change to the count, there is no change to
  the screen list and I can re-use the old computation. The count is
  stored in the variable screen_op_count.
 ***********************************************************************/
static int screen_op_count;

enum {
  GFXCLASS_LABEL, 
  GFXCLASS_LOGVIEW,
  GFXCLASS_SPINNER,
  GFXCLASS_PROGRESSBAR, 
  GFXCLASS_RADIO,   
  GFXCLASS_PANEL,
  GFXCLASS_BUTTON,
  GFXCLASS_TOGGLE,
  GFXCLASS_PIXGRID,
  GFXCLASS_CHECKBOX,
  GFXCLASS_DIRLIST,
  GFXCLASS_FILELIST,
  GFXCLASS_TEXTENTRY,
  GFXCLASS_SCROLLBAR,
  GFXCLASS_LIST,
  GFXCLASS_ICONLIST,
  GFXCLASS_SLIDER,
  GFXCLASS_DROPDOWN,
  GFXCLASS_CANVAS,
};

  // three strings. 0-normal 1-shift 2-symbol
typedef struct { int x,y,w,h; int func; char *str[3]; } vkey_t;

typedef struct
{
  int nkeys, nrows;
  vkey_t *keys;
  int state;
  int x,y,w,h;
  uint32_t *backstore;
  int pressed;
  int vs; // vertical spacing
} vkb_t;

#define OBJ_VKB    1

static unsigned int *checkbox_ok, *checkbox_nok;



static unsigned int  gfx_checkbox_width, gfx_checkbox_height;



// Reserved widget numbers
enum { 
     // confirmation dialog buttons
  GUIRaskB1=GFXEV_LAST, GUIRaskB2, GUIRaskB3, GUIRaskB4, GUIRaskB5,
  GUIRaskB6, GUIRaskB7, GUIRaskB8,
     // file selector widgets
  GUIRfilD1, GUIRfilD2, GUIRfilD3, GUIRfilD4, GUIRfilD5, GUIRfilD6,
  GUIRfilD7, GUIRfilD8, GUIRfilD9, GUIRfilD10, 
  GUIRfilHOME, GUIRfilREFRESH, GUIRfilOK, GUIRfilCANCEL, GUIRfilLIST,
  GUIRfilNAME,
     };


typedef struct
{
  const unsigned char *data;
  int data_length;
  int min_dpi; // minimum dpi value for this image
} gui_iconimage_t;

typedef struct
{
  int w, h;
  uint32_t *data;
  int n_images;
  gui_iconimage_t *images;
} gui_icon_t;                      


