/***********************************************************************
 *                                                                     *
 *                           WIDGET METHODS                            *
 *                                                                     *
 ***********************************************************************
 The following functions are common to many widgets. Normally, each
 function should be present in a class object, but I didn't do that 
 at the start. For now, some widgets use their function fields and others
 need to be called from the switch statements below. For example, when
 a widget needs to be destroyed, we call widget_destroy() instead of
 simply assuming that the destroy function in the widget object is valid.
 widget_destroy() does  the necessary check and if there is no such
 function, it calls the correct function based on the type of the widget.
 ***********************************************************************/
static void widget_destroy(widget_t* W)
{
  if (!W) return ;
  if (W->destroy) 
   { (*W->destroy)(W); widgets[W->number]= NULL; free(W); return ; }
  switch(W->klass)
  {
  case GFXCLASS_CANVAS: 
    { gui_canvas_t *CC= (void*) W;
      if (CC->buffer) free(CC->buffer);
      if (CC->scrollh) { scrollbar_destroy(CC->scrollh); free(CC->scrollh); }
      if (CC->scrollv) { scrollbar_destroy(CC->scrollv); free(CC->scrollv); }
      break; }
  case GFXCLASS_SCROLLBAR: scrollbar_destroy((scrollbar_t*)W); break;
  case GFXCLASS_TEXTENTRY: textentry_free((text_entry_t*)W); break;
  case GFXCLASS_SPINNER:  // #W.SPINNER
    gui_spinner_destroy((gui_spinner_t*) W);
    break;
  case GFXCLASS_RADIO:
    { radio_t *R= (void*) W;
      if (R->text) free(R->text);
      remove_click_handler(R->click); break; }
  case GFXCLASS_PIXGRID: 
    { gui_pixgrid_t *G;
      G= (gui_pixgrid_t*) W;
      remove_click_handler(G->click_handler);
      break; }
  default: break;
  }
  widgets[W->number]= NULL;
  free(W);
}

static void widget_move(widget_t *W,int x,int y,int w,int h)
{
  if (!W) return ;
  if (W->position) { (*W->position)(W,x,y,w,h); return ; }
  switch(W->klass)
  {
  case GFXCLASS_LABEL: gui_label_set_position((gui_label_t*)W, x,y,w,h); break; 
  case GFXCLASS_SPINNER: // #W.SPINNER
     gui_spinner_set_position((gui_spinner_t*)W, x,y,w,h);
     break; 
  case GFXCLASS_RADIO:   // #W.RADIO
     { radio_t *R; R= (void *)W;
       W->x= x; W->y= y; W->w= w; W->h= h;
       if (R->click!=-1) set_click_handler(R->click, x,y,w,h);
       radio_draw((radio_t*)W); break; }
  case GFXCLASS_PANEL: panel_position((panel_t *)W,x,y,w,h); break;
  case GFXCLASS_BUTTON:
  case GFXCLASS_TOGGLE:
     button_set_position( (gui_button_t*) W, x,y,w,h);
     break;
  case GFXCLASS_PIXGRID:
     { gui_pixgrid_t *G;
       G= (gui_pixgrid_t*) W;
       W->x= x;
       W->y= y;
       W->w= w;
       W->h= h;
       set_click_handler(G->click_handler,x,y,w,h);
       break; }
  case GFXCLASS_CHECKBOX:
     { checkbox_t *R; R= (void *)W;
       W->x= x; W->y= y; W->w= w; W->h= h;
       if (R->click!=-1) set_click_handler(R->click, x,y,w,h);
       checkbox_draw((checkbox_t*)W);
       break; }
  case GFXCLASS_TEXTENTRY: textentry_position((text_entry_t*)W,x,y,w,h); break;
  case GFXCLASS_LIST:
       { gui_list_set_position((gui_list_t*)W, x,y,w,h); break; }
  case GFXCLASS_SCROLLBAR: scrollbar_position((scrollbar_t*)W, x,y,w,h); break;
 case GFXCLASS_DROPDOWN: dropdown_position((dropdown_t*)W,x,y,w,h); break;
  default: break;
  }
}

static void widget_show(widget_t *W, int show)
{
  if (!W) return ;
  if (show) show= 1;

  if (W->show)
  {
    if (W->shown==show) return ;
    W->shown= show;
    if (W->shown) { (*W->show)(W); }
             else { if (W->hide) (*W->hide)(W); }
    return ;
  }

  switch(W->klass)
  {
  case GFXCLASS_CANVAS: 
    { gui_canvas_t *CC= (void*) W;
      if (CC->enabled==show) break;
      CC->enabled= show;
      if (CC->enabled)
      {
        gui_canvas_draw_all(CC);
        gui_canvas_update_click_handlers(CC);
      }
      break; } 
  case GFXCLASS_DROPDOWN: dropdown_show((dropdown_t*)W,show); break;
  case GFXCLASS_SCROLLBAR: scrollbar_show((scrollbar_t*)W, show); break;
  case GFXCLASS_LIST: gfx_list_show((gui_list_t*) W,show); break;
  case GFXCLASS_CHECKBOX: checkbox_show((checkbox_t*) W,show); break;
  case GFXCLASS_SPINNER: gui_spinner_show((gui_spinner_t*) W,show); break;
  case GFXCLASS_RADIO: radio_show((radio_t*) W,show); break;
  case GFXCLASS_PANEL: panel_show((panel_t*)W,show); break;
  case GFXCLASS_BUTTON:
  case GFXCLASS_TOGGLE: gui_button_show((gui_button_t*) W, show); break;
  case GFXCLASS_PIXGRID:
    { gui_pixgrid_t *B;
      B= (gui_pixgrid_t*) W;
      if (B->enabled==show) break ;
      B->enabled= show;
      break; }
  default: break;
  }
}

static void widget_set_text(widget_t *W, char *str)
{
  if (!W) return ;
  switch(W->klass)
  {
  case GFXCLASS_LABEL: gui_label_set_text( (gui_label_t*) W,str); break;
  case GFXCLASS_TOGGLE:
  case GFXCLASS_BUTTON: gui_button_set_text((gui_button_t*) W, str); break;
  case GFXCLASS_TEXTENTRY: text_entry_set((text_entry_t*)W,str); break;
  default: break;
  }
}

static char* widget_get_text(widget_t *W)
{
  char *str;
  if (!W) return strdup("");
  switch(W->klass)
  {
  case GFXCLASS_TEXTENTRY: str= textentry_get_text((text_entry_t*)W); break;
  default: str= strdup("");
  }
  return str;
}

static void widget_set_state(widget_t *W, int state)
{
  if (!W) return ;
  switch(W->klass)
  {
  case GFXCLASS_CHECKBOX: checkbox_set((checkbox_t *)W, state); break;
  case GFXCLASS_RADIO:
    { radio_t *R; R= (void*) W;
      if (R->state==state) break;
      if (R->state) { R->state= 0; radio_draw_button(R,R->state,1,1); }
      else radio_set(R);
      break; }
  case GFXCLASS_TOGGLE:
    { gui_button_t *B;
     B= (gui_button_t*) W;
     if (state) state= 1;
     button_set_state(B,state);
     break; }
  }
}

static int widget_get_state(widget_t *W)
{
  int result;
  if (!W) return 0;
  result= 0;
  switch(W->klass)
  {
  case GFXCLASS_RADIO: { radio_t *R; R= (void*) W; result= R->state; break; }
  case GFXCLASS_TOGGLE:
    { gui_button_t *B;
    B= (gui_button_t*) W;
    result= B->state;
    if (result) result= 1;
    break; }
  case GFXCLASS_CHECKBOX:
    { checkbox_t *R; R= (void*) W;
      result= R->state; break; }
  }
  return result;
}

static int widget_get_selected(widget_t *W)
{
  int result;
  if (!W) return -1;
  result= -1;
  switch(W->klass)
  {
  case GFXCLASS_ICONLIST:
    { gui_iconlist_t *IC= (gui_iconlist_t*)W;
      result= IC->sel; }
    break;
  case GFXCLASS_LIST: result= gfx_list_selected((gui_list_t *)W); break;
  case GFXCLASS_RADIO:
    { radio_t *R; R= (void*) W;
      if (R->group==-1) return -1;
      return radio_group_get_selected(R->group); }
case GFXCLASS_DROPDOWN: 
  {
    dropdown_t* DD;
    DD= (dropdown_t *) W;
    result= DD->selected;
    break;
  }
  }
  return result;
}

static void widget_set_selected(widget_t *W,int sel)
{
  if (!W) return ;
  switch(W->klass)
  {
  case GFXCLASS_PANEL: panel_activate((panel_t *)W,sel); break;
case GFXCLASS_DROPDOWN: gfx_dropdown_select((dropdown_t*)W, sel); break;
  case GFXCLASS_ICONLIST:
    gui_iconlist_select_item((gui_iconlist_t*)W, sel);
    break;
  case GFXCLASS_LIST: gfx_list_select((gui_list_t*)W,sel); break;
  }
}


static void widget_set_fgbg(widget_t *W,int fg,int bg)
{
  // It seems like only labels have this method..
  if (!W) return ;
  switch(W->klass)
  {
  // #W.LABEL
case GFXCLASS_LABEL: label_set_fgbg((gui_label_t*) W, fg, bg); break;
  }
}


static void widget_insert
  (widget_t *W,int pos, char *text, void *user, uint32_t *icon)
{
  if (!W) return ;
  switch(W->klass)
  {
case GFXCLASS_DROPDOWN: dropdown_insert((dropdown_t*)W, pos, text, user); break;
case GFXCLASS_LIST:
  gfx_list_insert((gui_list_t*) W, pos, text, user, icon);
  break;
  }
}

static void widget_remove(widget_t *W,int pos, void **Ruser)
{
  if (!W) return ;
  switch(W->klass)
  {
 case GFXCLASS_LIST: gfx_list_delete((gui_list_t*)W,pos,Ruser); break;
  case GFXCLASS_ICONLIST:
    gui_iconlist_remove((gui_iconlist_t*)W, pos, Ruser);
    break;
case GFXCLASS_DROPDOWN: dropdown_remove((dropdown_t*)W,pos,Ruser); break;
  }
}

static void widget_clear(widget_t *W)
{
  if (!W) return ;
  switch(W->klass)
  {
  case GFXCLASS_LIST: gfx_list_clear((gui_list_t*) W); break;
  case GFXCLASS_ICONLIST:
    gui_iconlist_clear((gui_iconlist_t*) W);
    break;
  case GFXCLASS_DROPDOWN: gui_dropdown_clear((dropdown_t*) W); break;
  }
}

void gfx_set_state(int wno, int state)
{
  widget_set_state(get_widget(wno),state);
}

int  gfx_get_state(int wno)
{
  return widget_get_state(get_widget(wno));
}

int  gfx_get_selected(int wno)
{
  return widget_get_selected(get_widget(wno));
}

void gfx_set_selected(int wno,int sel)
{
  widget_set_selected(get_widget(wno),sel);
}

void gfx_set_text(int wno,char *str)
{
  widget_set_text(get_widget(wno),str);
}

void gfx_set_fgbg(int wno, int fg, int bg)
{
  widget_set_fgbg(get_widget(wno),fg,bg);
}

void gfx_insert(int wno,int pos, char *text,void *user,unsigned int* icon)
{
  widget_insert(get_widget(wno),pos,text,user,icon);
}

void gfx_remove(int wno, int pos, void **Ruser)
{
  widget_remove(get_widget(wno),pos, Ruser);
}

char* gfx_get_text(int wno)
{
  return widget_get_text(get_widget(wno));
}

void gfx_clear(int wno)
{
   widget_clear(get_widget(wno));
}


/***********************************************************************
 *                                                                     *
 *                              UTILITIES                              *
 *                                                                     *
 ***********************************************************************/
static void init_data()
{
  // #W.RADIO
  gfx_radio_width= RES(radio.on).w;
  gfx_radio_height= RES(radio.on).h;

  gfx_checkbox_width= RES(checkbox.ok).w;
  gfx_checkbox_height= RES(checkbox.ok).h;

  fonts_init();
  scrollbar_init_buttons(NULL);

  init_guibc_globals();
}

void gfx_geometry(unsigned int *W, unsigned int *H)
{
  if (W) *W= x_window_width;
  if (H) *H= x_window_height;
}

void gfx_set_sysbar_height(int hh) { gui_sysbar_height= hh; }
int gfx_get_dpi() { return getdpi(); }
void gfx_pause() { updates_on= 0; }
void gfx_resume() { updates_on= 1; }

static void gfx_update_all()
{
  gfx_update(0,0,x_window_width, x_window_height);
}

/***********************************************************************
 *                                                                     *
 *                            CODE FOOTER                              *
 *                                                                     *
 ***********************************************************************/
void *__gfx_library__unused_symbol_holder__[]= 
  {
// Avoid unused warnings
   (void*) gui_canvas_new, (void*) gui_spinner, (void*) gui_logview, (void*) gfx_pixgrid, 
   (void*) dropdown_new, (void*) panel_new, (void*) gui_checkbox_new, (void*) radio_new,
   (void*) gui_slider_new, (void*) womap_destroy,
// imported utilities that could be removed
   (void*) buffer_append, 
   (void*) path_child, (void*) path_relative, (void*) path_dir,
   (void*) gui_iconlist_new,
   (void*) graypix_sub, 
   (void*) clip_rrect,
   (void*) obj_put, (void*) obj_get, (void*) obj_remove,
   (void*) gui_progressbar_new,
   (void*) gui_filelist_new,
   (void*) have_user_timer, (void*) elapsed_user_timer,
  };
