
/***********************************************************************
 *                                                                     *
 *                         BYTECODE INTERFACE                          *
 *                                                                     *
 ***********************************************************************
   In order for remote GUIs to work, I need to encode the GUI as data
   rather than code. This data will simply be bytecode.
   
   The execution environment will be a stack machine. Each instruction
   is one byte, with some optional data following it.
 ***********************************************************************/

/** BEGIN: BYTECODE  DEFINITIONS **/
#define GUII_C4          0x01
#define GUII_C3          0x02
#define GUII_C2          0x03
#define GUII_C1          0x04
#define GUII_ADD         0x05
#define GUII_SUB         0x06
#define GUII_MUL         0x07
#define GUII_DIV         0x08
#define GUII_MOD         0x09
#define GUII_STOREINT    0x0A
#define GUII_LOADINT     0x0B
#define GUII_LOCAL       0x0C
#define GUII_WLAY_X      0x0D
#define GUII_WLAY_Y      0x0E
#define GUII_WLAY_W      0x0F
#define GUII_WLAY_H      0x10
#define GUII_CF          0x11
#define GUII_GLOBAL      0x12
#define GUII_STR         0x13
#define GUII_CONSTRUCT   0x14
#define GUII_INTARRAY    0x15    
#define GUII_IPARAM       0x16
#define GUII_DONE        0x17
#define GUII_NOP         0x18
#define GUII_PPARAM      0x19

#define GUILV_WIDTH                   0
#define GUILV_HEIGHT                  1
#define n_predef_locals               2

#define GUIGV_FONT1HEIGHT             0
#define GUIGV_FONT2HEIGHT             1
#define GUIGV_FONT3HEIGHT             2
#define GUIGV_FONT4HEIGHT             3
#define GUIGV_FONT5HEIGHT             4
#define GUIGV_GUI_SCROLLBAR_WIDTH     5
#define GUIGV_GUI_SCROLLBAR_HEIGHT    6
#define GUIGV_GUI_SYSBAR_HEIGHT       7
// FIXME:  put radio width/height as well as checkbox width/height here.
#define n_global_vars                 8

//
// here is the numbering for the construction functions
//
enum { 
  GUIWC_BUTTON=1,     GUIWC_LABEL,              GUIWC_RADIO, 
  GUIWC_CHECKBOX,     GUIWC_PROGRESSBAR,        GUIWC_LIST, 
  GUIWC_PIXGRID,      GUIWC_TEXTENTRY,          GUIWC_PANEL,
  GUIWC_SPINNER,      GUIWC_DROPDOWN,           GUIWC_LOGVIEW,
  GUIWC_SLIDER,       GUIWC_CANVAS,             GUIWC_FILELIST,
  GUIWC_DIRLIST,      GUIWC_ICONLIST,           GUIWC_SCROLLBAR,
};

typedef struct 
{
  int length, n_locals;
  unsigned char *data; 
} bytecode_t;

/** END: BYTECODE DEFINITIONS **/

typedef struct
{
  int type;
  union { uint32_t uint;
	  int32_t  sint;
	  float    flo;
	  void    *ptr;
	} d;
} gui_cell_t;


static int guis_tos, guis_size;
static gui_cell_t *guis;

static int guibc_globals[n_global_vars];

#define GUIC_INT  1
#define GUIC_UNT  2
#define GUIC_FLO  3
#define GUIC_PTR  4

static void init_guibc_globals()
{
  guibc_globals[GUIGV_FONT1HEIGHT]=  font1height;
  guibc_globals[GUIGV_FONT2HEIGHT]=  font2height;
  guibc_globals[GUIGV_FONT3HEIGHT]=  font3height;
  guibc_globals[GUIGV_FONT4HEIGHT]=  font4height;
  guibc_globals[GUIGV_FONT5HEIGHT]=  font5height;
  guibc_globals[GUIGV_GUI_SCROLLBAR_WIDTH]= gui_scrollbar_width;
  guibc_globals[GUIGV_GUI_SCROLLBAR_HEIGHT]= gui_scrollbar_height;
  guibc_globals[GUIGV_GUI_SYSBAR_HEIGHT]= gui_sysbar_height;
}

static int guis_push(gui_cell_t *K)
{
  if (guis_tos==guis_size) return 1;
  guis[guis_tos]= *K;
  guis_tos++;
  return 0;
}


static int guis_pop(gui_cell_t *K)
{
  if (guis_tos==0) return 1;
  guis_tos--;
  *K= guis[guis_tos];
  return 0;
}

static void guis_init()
{
  if (!guis)
  {
    guis_size= 1000;
    guis= malloc(sizeof(guis[0])*guis_size); 
  }
  guis_tos= 0;
}

static void guis_Pint(int32_t k)
{
  gui_cell_t K;
  K.type= GUIC_INT;
  K.d.sint= k;
  guis_push(&K);
}

static void guis_Pptr(void *ptr)
{
  gui_cell_t K;
  K.type= GUIC_PTR;
  K.d.ptr= ptr;
  guis_push(&K);
}

static void* guis_Gptr()
{
  gui_cell_t K;
  if (guis_pop(&K)) return NULL;
  if (K.type==GUIC_PTR) return K.d.ptr;
  return NULL;
}

static void guis_Pflo(float k)
{
  gui_cell_t K;
  K.type= GUIC_FLO;
  K.d.flo= k;
  guis_push(&K);
}

static int32_t guis_Gint()
{ 
  gui_cell_t K;
  if (guis_pop(&K)) return 0;
  if (K.type==GUIC_FLO) return K.d.flo;
  return K.d.sint;
}

static float guis_Gflo()
{ 
  gui_cell_t K;
  if (guis_pop(&K)) return 0;
  if (K.type==GUIC_FLO) return K.d.flo;
  return K.d.sint;
}

static void construct_widget(int widget_no, int cons_func)
{
  widget_t *W;

  int i1, i2, i3, i4, i5, i6;
  float f1, f2, f3;
  char *s1;
  void *p1;
  
  W= NULL;
  switch(cons_func)
  {
  case GUIWC_DIRLIST: W=  gui_dirlist_new(NULL,NULL,NULL); break; 
  case GUIWC_FILELIST:
    i1= guis_Gint(); 
    i2= guis_Gint(); 
    W= gui_filelist_new(i1,i2,NULL,NULL,NULL);
    break;
  case GUIWC_BUTTON:
    s1= guis_Gptr();
    i1= guis_Gint(); 
    i2= guis_Gint(); 
    i3= guis_Gint();
    W= gfx_button(s1, i1, i2, i3);
    break;
  case GUIWC_LABEL:
    i1= guis_Gint();
    i2= guis_Gint();
    i3= guis_Gint();
    i4= guis_Gint();
    s1= guis_Gptr();
    W= gfx_label(i1,i2,i3,i4,s1);
    break;
  case GUIWC_RADIO:
    s1= guis_Gptr();
    i1= guis_Gint();
    W= radio_new(s1, i1);
    break;
  case GUIWC_CHECKBOX:
    s1= guis_Gptr();
    W= gui_checkbox_new(s1,NULL,NULL);
    break;
  case GUIWC_PROGRESSBAR:
    i1= guis_Gint();
    i2= guis_Gint();
    i3= guis_Gint();
    i4= guis_Gint();
    W= gui_progressbar_new(i1,i2,i3,i4);
    break;
  case GUIWC_LIST: W= gui_list_new(0,0,NULL,NULL,NULL); break;
  case GUIWC_PIXGRID: W= gfx_pixgrid(); break;
  case GUIWC_TEXTENTRY: 
    s1= guis_Gptr();
    W= text_entry_new(s1);
    break;
  case GUIWC_PANEL:
    p1= guis_Gptr();
    W= panel_new(p1);
    break;
  case GUIWC_SPINNER: W= gui_spinner('I',3); break;
  case GUIWC_DROPDOWN: W= dropdown_new(NULL,NULL); break;
  case GUIWC_LOGVIEW: W= gui_logview(); break;
  case GUIWC_SLIDER:
    i1= guis_Gint();
    W= gui_slider_new(i1);
    break;
  case GUIWC_CANVAS:
    i1= guis_Gint();
    i2= guis_Gint();
    W= gui_canvas_new(i1,i2);
    break;
  case GUIWC_ICONLIST:
    i1= guis_Gint(); i2= guis_Gint(); i3= guis_Gint();
    i4= guis_Gint(); i5= guis_Gint(); i6= guis_Gint();
    W= gui_iconlist_new(i1,i2,i3,i4,i5,i6);
    break;
  case GUIWC_SCROLLBAR:
    i1= guis_Gint(); f1= guis_Gflo();
    f2= guis_Gflo(); f3= guis_Gflo();
    W= scrollbar_new(i1,f1,f2,f3,0,NULL,NULL);
    break;
  }
  if (!W) return ;
  widgets[widget_no]= W;
  W->number= widget_no;
}

static void gui_run_bytecode
  (const unsigned char *code, int length,int *locals,int *layout)
{
  int widget_no, cons_func;
  int pc;
  gui_cell_t A1, A2;
  unsigned char C;
  char *Z;
  int *AP;
  int i1, i2;
  float f1, f2;
  int iV, *iA, i;

  guis_init();
  pc= 0;

next:
  if (pc>=length) return ;
  C= code[pc++];
  if (C&0x80)
  {
    guis_Pint(C&0x7f);
    goto next;
  }
  iV= 0;
  switch(C)
  {
  case GUII_CF:  
  case GUII_C4: iV= code[pc++];
  case GUII_C3: iV= (iV<<8) | code[pc++];
  case GUII_C2: iV= (iV<<8) | code[pc++];
  case GUII_C1: iV= (iV<<8) | code[pc++]; 
     if (C==GUII_CF)
     { union { float F; int32_t I; } D;
       D.I= iV; guis_Pflo(D.F); }
     else guis_Pint(iV); 
     goto next;
  case GUII_STOREINT: if (AP) *AP= guis_Gint(); goto next;
  case GUII_LOADINT: if (AP) guis_Pint(*AP); goto next;
  case GUII_LOCAL: AP= locals + guis_Gint(); goto next;
  case GUII_GLOBAL: AP= guibc_globals + guis_Gint(); goto next;
  case GUII_WLAY_X: AP= layout + 4*guis_Gint() + 0; goto next;
  case GUII_WLAY_Y: AP= layout + 4*guis_Gint() + 1; goto next;
  case GUII_WLAY_W: AP= layout + 4*guis_Gint() + 2; goto next;
  case GUII_WLAY_H: AP= layout + 4*guis_Gint() + 3; goto next;
  case GUII_STR: 
       iV= guis_Gint();
       Z= (char*) code+pc;
       pc+= iV;
       guis_Pptr(Z);
       break;
  case GUII_ADD:
  case GUII_SUB:       
  case GUII_MUL:      
  case GUII_DIV:     
  case GUII_MOD:  
     goto arith;
  case GUII_CONSTRUCT:
     widget_no= guis_Gint();
     cons_func= guis_Gint();
     construct_widget(widget_no, cons_func);
     break;
 case GUII_INTARRAY:
     iV= guis_Gint();
     iA= NULL;
     if (iV) iA= malloc(sizeof(int)*iV);
     for(i=0;i<iV;i++) iA[iV-1-i]= guis_Gint();
     guis_Pptr(iA);
     break; 
  case GUII_NOP:
     break;
  }
  goto next;
arith:
  guis_pop(&A2);
  guis_pop(&A1);
  if (A1.type==GUIC_FLO || A2.type==GUIC_FLO) 
  {
    if (A1.type==GUIC_FLO) f1= A1.d.flo;
                      else f1= A1.d.sint;
    if (A2.type==GUIC_FLO) f2= A2.d.flo;
                      else f2= A2.d.sint;
    switch(C)
    {
    case GUII_ADD: f1= f1 + f2; break;
    case GUII_SUB: f1= f1 - f2; break;
    case GUII_MUL: f1= f1 * f2; break;
    case GUII_DIV: if (f2!=0) f1= f1 / f2; break;
    case GUII_MOD: i1= f1; i2= f2; goto intarith;
    }
    guis_Pflo(f1);
    goto next;
  }
  i1= A1.d.sint;
  i2= A2.d.sint;
intarith:
   switch(C)
   {
    case GUII_ADD: i1= i1 + i2; break;
    case GUII_SUB: i1= i1 - i2; break;
    case GUII_MUL: i1= i1 * i2; break;
    case GUII_DIV: if (i2!=0) i1= i1 / i2; break;
    case GUII_MOD: if (i2!=0) i1= i1 % i2; break;
   } 
   guis_Pint(i1);
   goto next;
}


static void gui_layout_widgets
  (int WIDTH, int HEIGHT, int nlocals, int length,
   const unsigned char *code, int *Wcoor)
{
  int *locals;
  locals= malloc(sizeof(*locals)*nlocals);
  locals[GUILV_WIDTH]= WIDTH;
  locals[GUILV_HEIGHT]= HEIGHT;
  gui_run_bytecode(code, length, locals, Wcoor);                
  free(locals);
}

static void gui_compute_screen_dimensions
  (int parent_w,int parent_h,
   int n_locals, int length, const unsigned char *code,
   int *Scoor)
{
  int *locals;
  locals= malloc(sizeof(*locals)*n_locals);
  locals[GUILV_WIDTH]= parent_w;
  locals[GUILV_HEIGHT]= parent_h;
  Scoor[0]= Scoor[1]= Scoor[2]= Scoor[3]= -1;
  gui_run_bytecode(code, length, locals, Scoor);
  free(locals);
}

static void gui_construct_widgets
  (const unsigned char *code, int length)
{
  int locals[1];
  locals[0]= 0;
  gui_run_bytecode(code, length, locals, locals);
}

//
// this is similar to run_bytecode, but recognizes much less instructions.
// it's used for parsing metadata. It leaves the requested type of data
// at the top of the stack. The returned value is the rest of the code.
//
static const unsigned char* guibc_get_param
  (const unsigned char *code, int type)
{
  int pc;
  unsigned char C;
  int iV,i,*iA;

  guis_init();

  pc= 0;
next:
  C= code[pc++];
  if (C&0x80)
  {
    guis_Pint(C&0x7f);
    goto next;
  }
  iV= 0;
  switch(C)
  {
  case GUII_C4: iV= code[pc++];
  case GUII_C3: iV= (iV<<8) | code[pc++];
  case GUII_C2: iV= (iV<<8) | code[pc++];
  case GUII_C1: iV= (iV<<8) | code[pc++]; 
     guis_Pint(iV); 
     break;
  case GUII_IPARAM:
  case GUII_PPARAM:
     if (type==C) return code+pc;
     break;
 case GUII_INTARRAY:
     iV= guis_Gint();
     iA= NULL;
     if (iV) iA= malloc(sizeof(int)*iV);
     for(i=0;i<iV;i++) iA[iV-1-i]= guis_Gint();
     guis_Pptr(iA);
     break; 
  }
  goto next;
}

static int guibc_get_iparam(const unsigned char **Rcode)
{
  *Rcode= guibc_get_param(*Rcode, GUII_IPARAM);
  return guis_Gint();
}

static void* guibc_get_pparam(const unsigned char **Rcode)
{
  *Rcode= guibc_get_param(*Rcode, GUII_PPARAM);
  return guis_Gptr();
}

static void guibc_get_code
  (const unsigned char **descr, int *length, const unsigned char **code)
{
  *length= guibc_get_iparam(descr);
  *code= *descr;
  *descr += *length;
}

static void gui_screen_start(const unsigned char *sdef)
{
  int n_widgets, n_direct_widgets, n_subs, n_locals; 
  int *widget_ids;
  int Scoor[4];
  int *Wcoor;
  int pindex, start, end;
  int z, i, j;

  int code_length;
  const unsigned char *code;
 
  n_widgets= guibc_get_iparam(&sdef);
  n_direct_widgets= guibc_get_iparam(&sdef);
  widget_ids= guibc_get_pparam(&sdef);

  //
  // The screen layout code. See below comment about how this could
  // be used to handle window resize evets.
  //
  n_locals= guibc_get_iparam(&sdef);
  guibc_get_code(&sdef, &code_length, &code);

  gui_compute_screen_dimensions(screen?screen->w:x_window_width,
                                screen?screen->h:x_window_height,
                      n_locals, code_length, code, Scoor);
  enter_screen(n_widgets, widget_ids, Scoor);

  // constructor code
  guibc_get_code(&sdef, &code_length, &code);
  gui_construct_widgets(code, code_length);

  //
  // now, laying out the stuff
  // At this point, we could store the definition object within the
  // screen struct. This would allow us to re-layout widgets when
  // the window gets resized. I'm not going to do that for now.
  //
  //
  // This is the main layout code.
  //
  n_locals= guibc_get_iparam(&sdef);
  guibc_get_code(&sdef, &code_length, &code);

  Wcoor= malloc(sizeof(int)*n_widgets*4);
  gui_layout_widgets(Scoor[2],Scoor[3], n_locals, code_length, code, Wcoor);

  n_subs= guibc_get_iparam(&sdef);

  for(z=0;z<n_subs;z++)
  {
    pindex= guibc_get_iparam(&sdef);
    start= guibc_get_iparam(&sdef);
    end= guibc_get_iparam(&sdef);
    n_locals= guibc_get_iparam(&sdef);
    guibc_get_code(&sdef, &code_length, &code);
    
    gui_layout_widgets(Wcoor[pindex*4+2], Wcoor[pindex*4+3],
                       n_locals, code_length, code, Wcoor+4*start);
    for(j=start;j<end;j++)
    {
       Wcoor[4*j]+= Wcoor[4*pindex];
       Wcoor[4*j+1] += Wcoor[4*pindex+1];
    }
  }
  for(i=0;i<n_widgets;i++)
  {
    Wcoor[4*i]+= Scoor[0];
    Wcoor[4*i+1]+= Scoor[1];
  }

  for(i=0;i<n_widgets;i++)
     widget_move(widgets[widget_ids[i]], 
                               Wcoor[4*i], Wcoor[4*i+1],
                               Wcoor[4*i+2], Wcoor[4*i+3]);
  for(i=0;i<n_direct_widgets;i++)
     widget_show(widgets[widget_ids[i]],1);
  free(Wcoor);
  gfx_update_all();        
}

static int gui_parse_int(const unsigned char **Rdata)
{
  int R;
  const unsigned char *D;
  D= *Rdata;
  R= (D[0]<<24) | (D[1]<<16) | (D[2]<<8) | D[3];
  *Rdata += 4;
  return R;
}

static int n_screen_defs;

typedef struct
{
  const unsigned char *data;
  int size;
} screen_def_t;

static screen_def_t *screen_defs;

static const unsigned char* gui_parse_icons(const unsigned char *data);

void gfx_data(const unsigned char *mdata)
{
  int i;
  int ns;

  mdata= gui_parse_icons(mdata);

  first_widget_no= gui_parse_int(&mdata);
  last_widget_no= gui_parse_int(&mdata);
  n_screen_defs= gui_parse_int(&mdata);
  
  widgets= calloc(sizeof(widgets[0]), last_widget_no+1);
  screen_defs= malloc(sizeof(screen_defs[0])*n_screen_defs);

  for(i=0;i<n_screen_defs;i++)
  {
    screen_defs[i].size= gui_parse_int(&mdata);
    screen_defs[i].data= mdata; 
    mdata+= screen_defs[i].size;
  }
}

void gfx_screen(int num)
{
  if (num<0 || num>=n_screen_defs) return ;
  gui_screen_start(screen_defs[num].data);
}
