/***********************************************************************
 *                                                                     *
 *                             User Icons                              *
 *                                                                     *
 ***********************************************************************
  User icons are simply stored in an array. Indices to this array are
  mentioned in either widget creation functions or list insert functions.

  Each icon is a set of images, one for class of screen resolutions. 
  The range of valid resolutions for each image is determined by the
  user, given as a numerical value. 

  Once an icon is loaded, it's converted to user image format and
  then used from that raw data.

  In the module description, icons are represented as follows:
    n_icons  : int
    <for each icon>
       n_images : int
       <for each image>
         min_dpi : int
         data_length: int
         data:  bytes
 ***********************************************************************/


static gui_icon_t **gui_icons;
static int gui_n_icons;

static unsigned const char *gui_parse_iconimage
   (const unsigned char *src, gui_iconimage_t *icon)
{
  icon->min_dpi= gui_parse_int(&src);
  icon->data_length= gui_parse_int(&src);
  icon->data= src;
  return src+icon->data_length;
}

static unsigned const char *gui_parse_icon
   (const unsigned char *src, gui_icon_t *icon)
{
  int i;
  icon->n_images= gui_parse_int(&src);
  icon->images= malloc(sizeof(icon->images[0])*icon->n_images);
  for(i=0;i<icon->n_images;i++)
    src= gui_parse_iconimage(src, icon->images+i);
  return src;
}

static unsigned const char *gui_parse_icons(const unsigned char *src)
{
  int i;
  gui_n_icons= gui_parse_int(&src);
  if (!gui_n_icons) return src;
  gui_icons= malloc(sizeof(gui_icons[0])*gui_n_icons);
  for(i=0;i<gui_n_icons;i++)
  {
    gui_icons[i]= calloc(1,sizeof(gui_icons[i][0]));
    src= gui_parse_icon(src, gui_icons[i]);
  }
  return src;
}

static void gui_decode_icon(gui_icon_t *icon)
{
  int i,j;
  unsigned W, H;
  int length;
  unsigned char *byte_data;
  gui_iconimage_t *img;
  int dpi;

  dpi= getdpi();
  img= icon->images+0;
  for(i=1;i<icon->n_images;i++)
   if (dpi>icon->images[i].min_dpi)
     img= icon->images+i;
  //
  // we know the following won't fail because we did the same thing
  // while compiling the module and got no errors.
  //
  lodepng_decode32(&byte_data, &W, &H, img->data, img->data_length);
  icon->w= W;
  icon->h= H;
  icon->data= malloc(sizeof(icon->data[0])*W*H);
  length= W*H;
  for(i=0,j=0;i<length;i++,j+=4)
    icon->data[i]=   (byte_data[j+3]<<24) | (byte_data[j]<<16) 
                    | (byte_data[j+1]<<8) | (byte_data[j+2]) ;
  free(byte_data);
}

//
// icon_no is from 1 to gui_n_icons
// 0 is used for no-icon situations
//
static gui_icon_t *gui_get_icon(int icon_no)
{
  gui_icon_t *icon;
  if (icon_no<1 || icon_no>gui_n_icons) return NULL;
  icon_no--;
  icon= gui_icons[icon_no];
  if (!icon->data) gui_decode_icon(icon);
  return icon;
}
